# -*- coding: utf-8 -*-
"""
GRN Motif Binding Site Inference
--------------------------------

This module allows to infer Gene Regulatory Networks based on Transcription
Factor binding sites motifs. This module relies on the `iRegulon`_ dataset,
which has also been used for `i-cistarget`_.

Example:
    Test the example by running this file::

        $ python motifsearch.py

.. _iRegulon:
   http://iregulon.aertslab.org/collections.html#motifcolldownload

.. _i-cistarget:
    https://github.com/aertslab/pySCENIC 
"""

from Bio import SeqIO
from Bio import motifs
from os.path import join
from os import listdir
import pandas as pd
import numpy as np
from tqdm import tqdm
from Bio.Alphabet import IUPAC
import multiprocessing as mp

__author__ = "Sergio Peignier"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.0.1"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"

def load_motifs(folder):
    """
    Load 10K motifs (position weight matrices PWM) from iRegulon
    Args:
        folder: path to the motif files

    Returns:
        list: [Bio.motifs.jaspar.Motif] list of motifs
    """
    pwds_total = []
    for f in tqdm(listdir(folder)):
        if f[0] != ".":
            f_path = join(folder,f)
            pwd = motifs.parse(open(f_path,"r"), "jaspar")[0]
            pwds_total.append(pwd)
            #print(pwd.name)
    return(pwds_total)

def motif2pssm(motif, alphabet):
    """
    convert motif to PSSM (position-specific scoring matrix) distribution
    Args:
        motif (Bio.motifs.jaspar.Motif): biopython motif (PWM)
        alphabet (dict): letter to frequency in the genome

    Returns:
        dict: {<original motif>, <pssm>, <distribution>}
    """
    # Add pseudocounts to prevent zero probabilities
    ppm = motif.counts.normalize(pseudocounts=alphabet)
    # Set the background (species biases)
    pssm = ppm.log_odds(background=alphabet)
    # Compute the distribution of PWM scores
    distribution = pssm.distribution(background=alphabet, precision=10**3)
    return(motif.name, {"motif":motif,"pssm":pssm,"distribution":distribution})

def motifs2pssms(motifs, alphabet):
    """
    Converts each motif in pssm
    Args:
        motif list : list of biopython motif (PWM)
        alphabet (dict): letter to frequency in the genome

    Returns:
        dict: dictionnary motifID -> {<original motif>, <pssm>, <distribution>}
    """
    pssms = {}
    for motif in tqdm(motifs):
        motif_name, pssm = motif2pssm(motif, alphabet)
        pssms[motif_name] = pssm
    return(pssms)

def max_aggregation(scores):
    """
    Computes the maximal score for a list of scores, NaN are set to 0

    Args:
        scores (list): list of scores

    Returns
        float : maximal score in sequence

    Examples:
        >>> max_aggregation([1,2,np.nan])
        2
    """
    return(np.nan_to_num(scores).max())

def score_sequence_with_pssm(sequence, pssm, f_aggregation=None):
    """
    Score the sequence positions w.r.t. a PSSM and aggregates the resulting
    vector to output a single score

    Args:
        sequence (Bio.Seq.Seq): Biopython sequence
        pssm (Bio.motifs.matrix.PositionSpecificScoringMatrix): PSSM
        f_aggregation (func): Aggregation function

    Returns:
        float or numpy.array: if f_aggregation is not None, a single float score
        is returned, otherwise a numpy.array with the score for each location
        of the sequence is returned
    """
    sequence.alphabet = pssm.alphabet
    score = pssm.calculate(sequence)
    if f_aggregation is not None:
        score = f_aggregation(score)
    return(score)


def score_sequence_pssms_dict(sequence,
                              pssm_dic,
                              f_aggregation=None):
    """
    Score a sequence w.r.t. a dictionnary of PSSM

    Args:
        sequence (Bio.Seq.Seq): Biopython sequence
        pssm (dict): dict of Bio.motifs.matrix.PositionSpecificScoringMatrix
        f_aggregation (func): Aggregation function

    Returns:
        dict: PSSM key -> sequence score for the given PSSM
    """
    scores = {}
    for k in pssm_dic:
        scores[k] = score_sequence_with_pssm(sequence,
                                             pssm_dic[k],
                                             f_aggregation)
    return(scores)

def _score_sequence_pssms_dict(sequence_id,
                               sequence,
                               pssm_dict,
                               f_aggregation=None):
    """
    Applies score_sequence_pssms_dict, but also returns the sequence id

    Args:
        sequence_id (str): sequence id
        sequence (Bio.Seq.Seq): Biopython sequence
        pssm (dict): dict of Bio.motifs.matrix.PositionSpecificScoringMatrix
        f_aggregation (func): Aggregation function

    Returns:
        str: sequence id
        dict: PSSM key -> sequence score for the given PSSM
    """
    scores = score_sequence_pssms_dict(sequence,pssm_dict,f_aggregation)
    return(sequence_id,scores)

def score_sequences(genes_sequences,
                    pssm_distrib_dict,
                    f_aggregation=None,
                    n_jobs=1):
    """
    Scores sequences w.r.t. PSSM motifs

    Args:
        genes_sequences (dict): geneID -> contextual sequence
        pssm_distrib_dic (dict):
            motifID -> {<original motif>, <pssm>, <distribution>}
        f_aggregation (func): Aggregates the score along the sequence
        n_jobs (int): number of jobs to run in parallel

    Returns:
        dict: gene -> motif -> scores along sequence and context
    """

    def chunk_gene_sequences(genes_sequences, n=100):
        l = genes_sequences.items()
        l = list(l)
        for i in range(0, len(l), n):
            yield dict(l[i:i + n])
    # create a dictionnary of pssms only
    pssm_dict = {k:pssm_distrib_dict[k]["pssm"] for k in pssm_distrib_dict}
    # chunk the gene sequences dict
    sub_gene_sequences = list(chunk_gene_sequences(genes_sequences))
    # Set the number of jobs
    if n_jobs > mp.cpu_count():
        n_jobs = mp.cpu_count()
    pool = mp.Pool(n_jobs)
    # generate multiprocess tasks
    scores_mp = []
    for sub_sequences in tqdm(sub_gene_sequences):
        scores_mp +=  [pool.apply_async(_score_sequence_pssms_dict,
                                        args=(seq_id,seq,pssm_dict,f_aggregation))
                                        for seq_id,seq in sub_sequences.items()]
    pool.close()
    # Run multiprocess computations
    scores = []
    for p in tqdm(scores_mp):
        scores.append(p.get())
    scores = pd.DataFrame(dict(scores))
    return(scores)


def get_max_score(scores):
    """
    Get highest score in sequence
    Args:
        dict: gene -> motif -> scores along sequence and context

    Returns:
        dict: gene -> motif -> best score

    TODO:
        let the selection function be a parameter
    """
    max_score = {}
    for gene in tqdm(scores):
        max_score[gene] = {}
        for tf in scores[gene]:
            max_score[gene][tf] = np.nan_to_num(scores[gene][tf]).max()
    return(max_score)

def compute_threshold(pssm_distrib_dic, fpr=0.001):
    """
    Compute a likelihood threshold based on the distribution of the motif
    the false positive rate is set as a parameter
    Args:
        pssm_distrib_dic (dict):
            motifID -> {<original motif>, <pssm>, <distribution>}
        fpr (float): False Positive Rate

    Returns:
        dict: motif -> threshold
    """
    thresholds = {}
    for motif in pssm_distrib_dic:
        distribution =  pssm_distrib_dic[motif]["distribution"]
        thresholds[motif] = distribution.threshold_fpr(fpr)
        #.threshold_balanced(100)
        #threshold_fpr(0.0001)
        #.threshold_patser()
        #.threshold_fnr(0.5)
    return(thresholds)


if __name__ == '__main__':
    from aphids.genome import alphabet_acypi
    motifs = load_motifs()
    # for m in motifs:
    #     print(m)
    pssms = motifs2pssms(motifs, alphabet_acypi)
    # import pickle
    # pssms = pickle.load(open('/Users/sergiopeignier/Documents/libraries/grn/models/motifs/pssms_10k.pickle',"rb"))
    for pssm in pssms:
        print(pssm)
    from aphids.genome import load_aphidbase_gene_sequences
    genes_sequences = load_aphidbase_gene_sequences(delta_pre=3000,
                                                    delta_post=8000)
    scores_full = score_sequences(genes_sequences,pssms)
    scores_best = get_max_score(scores_full)
    scores_df = pd.DataFrame(scores_best)
    import matplotlib.pyplot as plt
    plt.imshow(scores_df.T,aspect="auto")
    plt.show()
    thresholds = compute_threshold(pssms)
    plt.imshow(scores_df.T > pd.Series(thresholds),aspect="auto")
    plt.show()



# -*- coding: utf-8 -*-
"""
Mapping TF list to motif2TF file
--------------------------------

This module allows to write a subcollection of interactions between
transcription factor binding sites (TFBSs) and transcription factors (TFs), based on
a list of TFs. The interactions in the subcollection will involve only the TFs in the list.

Example:
    Test the example by running this file::

        $ python mapping.py
"""

import pandas as pd 

__author__ = "Timothée Frouté and Sergio Peignier"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.0.1"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"

def write_ortho_motifs(motifs_file_path,ortho_gene_list_file_path,outfile="../mapped_motifs.tbl"):
    """
    Gets and writes a subcollection of motifs-TF interactions based on list of transcription factors. 

    Args:
        motifs_file_path (str): motif2tf file path from the cistarget databases 
        ortho_gene_list_file_path (str): name of the scores .feather file to write

    Examples:
        >>> ortho_gene_list_file_path = "../dros_ortho_tf_puc.txt"
        >>> motifs_file_path = "../motifs-v9-nr.flybase-m0.001-o0.0.tbl"
        >>> write_ortho_motifs(motifs_file_path,ortho_gene_list_file_path)
    """
    f = open(ortho_gene_list_file_path)
    ortho_gene_list = []
    for gene in f.readlines():
        ortho_gene_list.append(gene.strip())
    ortho_gene_list = list(dict.fromkeys(ortho_gene_list)) #gets rid of duplicates, if any
    total_motifs = pd.read_csv(motifs_file_path,sep="\t",low_memory=False)
    df = pd.DataFrame(columns=total_motifs.columns)
    for gene_name in ortho_gene_list:
        if gene_name in list(total_motifs["gene_name"]):
            df = df.append(total_motifs.loc[total_motifs["gene_name"] == gene_name],ignore_index=True)
    df.to_csv(outfile, sep="\t", index=0)
    return df

if __name__ == "__main__":
    id_list_file = "../example_list.txt"
    motifs_file = "../motifs-v9-nr.flybase-m0.001-o0.0.tbl"
    motifs = write_ortho_motifs(motifs_file,id_list_file)


# -*- coding: utf-8 -*-
"""
Creating and saving feather files
---------------------------------

This module allows to write the feather files (scores and ranks) 
obtained with the motifsearch module.

Example:
    Test the example by running this file::

        $ python create_feather.py

Todo:
    * compare with pyscenic creat_cisTarget_database: https://github.com/aertslab/create_cisTarget_databases

"""

import os,sys
import pickle
import pandas as pd
import pyarrow.feather as feather
from feather_creator.motifsearch import score_sequences, max_aggregation, motifs2pssms, load_motifs

__author__ = "Timothée Frouté and Sergio Peignier"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.0.1"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"

def test_feather_read_write():
    """
    Tests the writing and the reading of a .feather file using pyarrow 0.16.0 functions.
    """
    data = {'product_name': ['laptop', 'printer', 'tablet', 'desk', 'chair'],
    'price': [1200, 150, 300, 450, 200]
    }

    df = pd.DataFrame(data)

    writer = feather.FeatherWriter("./test.feather") 
    writer.write(df)

    test = feather.read_feather("test.feather")
    print(test)

def save_scores_feather(scores_df,scores_feather_file):
    """
    Feather file writer, allows writting a feather file, used to write the motifs scores dataframe.

    Args:
        scores_df (pandas.DataFrame): scores matrix obtained with the score_sequences function,
            where each row is a motif name and each column is a gene sequence name
        feather_file (str): name of the .feather file to write

    Examples:
        >>> from motifsearch import score_sequences
        >>> from motifsearch import max_aggregation
        >>> scores_best = score_sequences(genes_sequences,pssms,max_aggregation,n_jobs=8)
        >>> save_scores_feather(scores_best,"./scores_full_acypi_background.feather")
    """
    writer = feather.FeatherWriter(scores_feather_file) 
    scores_df.reset_index(level=0, inplace=True)
    writer.write(scores_df)

def load_scores_feather(scores_feather_file):
    """
    Feather file loader, loads a feather file, used to load the motifs scores dataframe.

    Args:
        scores_df (pandas.DataFrame): scores matrix obtained with the score_sequences function,
            where each row is a motif name and each column is a gene sequence name
        scores_feather_file (str): name of the .feather file to load

    Examples:
        >>> save_scores_feather(scores_best,"./scores_full_acypi_background.feather")
        >>> scores_df = load_scores_feather("scores_full_acypi_background.feather")
        >>> scores_df
                               ACYPI084998  ACYPI44835  ACYPI010106  ACYPI26530  ACYPI089001  ...  ACYPI085136  ACYPI000318  ACYPI30960  ACYPI24141  ACYPI006702
        elemento-AAAATGGCG     1.784074    0.000000     0.000000    0.000000     1.784074  ...     0.000000     0.000000    0.000000    0.000000     0.000000
    """
    scores_df = feather.read_feather(scores_feather_file)
    scores_df = scores_df.set_index("index")
    return scores_df

def save_ranks_feather(scores_df,ranks_feather_file): 
    """
    Feather file writer, allows the ranking and writting of a feather file, used to write the motifs ranks dataframe.

    Args:
        scores_df (pandas.DataFrame): scores matrix obtained with the score_sequences function,
            where each row is a motif name and each column is a gene sequence name
        ranks_feather_file (str): name of the .feather file to write

    Examples:
        >>> from motifsearch import score_sequences
        >>> from motifsearch import max_aggregation
        >>> scores_best = score_sequences(genes_sequences,pssms,max_aggregation,n_jobs=8)
        >>> save_ranks_feather(scores_df,"./ranks_full_acypi_background.feather")
        >>> rank_df = feather.read_feather("ranks_full_acypi_background.feather")
        >>> rank_df
                    features  ACYPI084998  ACYPI44835  ACYPI010106  ACYPI26530  ACYPI089001  ...  ACYPI065189  ACYPI085136  ACYPI000318  ACYPI30960  ACYPI24141  ACYPI006702
            0          0         3210       10749        10750       10751         3211  ...        36933        36934        36935       36936       36937        36938
            1          1         3360        3361        17718       15937        17719  ...        36935        36936        15936       36937       36938         2180
 
    """
    ranks_df = (-scores_df).rank(axis=1,method="first")
    ranks_df = ranks_df - 1
    ranks_df = ranks_df.astype(int)
    ranks_df.reset_index(level=0, inplace=True)
    cols = [c for c in ranks_df.columns]
    cols[0] = "features"
    ranks_df.columns = cols
    writer = feather.FeatherWriter(ranks_feather_file) 
    writer.write(ranks_df)

def full_process(genes_sequences, pssms, scores_feather_file, ranks_feather_file,output_dir = "../output", alphabet=False):
    """
    Scores sequences w.r.t. PSSM motifs, writes scores and ranks feather files.

    Args:
        genes_sequences (dict): geneID -> contextual sequence 
        pssms (str or dict):
            motif dir path, pickle path or 
            motifID -> {<original motif>, <pssm>, <distribution>}
        scores_feather_file (str): name of the scores .feather file to write
        ranks_feather_file (str): name of the ranks .feather file to write
        output_dir (str): name of the output directory
        alphabet (dict): letter to frequency in the genome

    Examples:
        >>> motifs = load_motifs("motifs")
        >>> pssms = motifs2pssms(motifs, alphabet_acypi)
        >>> genes_sequences = load_aphidbase_gene_sequences(delta_pre=2000,
                                                    delta_post=2000)
        >>> genes_sequences = {k:genes_sequences[k] for k in list(genes_sequences.keys())}                                            
        >>> full_process(genes_sequences, pssms,"./scores_full_acypi_background.feather", "./ranks_full_acypi_background.feather")
    """
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    if type(pssms) is str:
        if pssms.endswith(".pickle") and os.path.isfile(pssms):
            pssms = pickle.load(open(pssms,"rb"))
        
        elif os.path.isdir(pssms) and len(os.listdir(pssms)) != 0 and alphabet:
            pssms = motifs2pssms(pssms, alphabet)
    if type(pssms) is not dict:
        print("Please use motif dir, .pickle file or correct pssms dict format")
        return 0
    scores_best = score_sequences(genes_sequences,pssms,max_aggregation,n_jobs=8)
    save_scores_feather(scores_best,os.path.join(output_dir,scores_feather_file))
    scores_df = load_scores_feather(os.path.join(output_dir,scores_feather_file))
    save_ranks_feather(scores_df,os.path.join(output_dir,ranks_feather_file))

if __name__ == "__main__":

    from aphids.genome import load_aphidbase_gene_sequences
    from aphids.genome import alphabet_acypi
    motifs = load_motifs("../motifs")
    pssms = motifs2pssms(motifs, alphabet_acypi)

    genes_sequences = load_aphidbase_gene_sequences(delta_pre=2000,
                                                    delta_post=2000)
                                    
    genes_sequences = {k:genes_sequences[k] for k in list(genes_sequences.keys())}

    full_process(genes_sequences, pssms,"./scores_full_acypi_background.feather", "./ranks_full_acypi_background.feather")




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = []
    for package in f.readlines():
        requirements.append(package.strip())

setuptools.setup(
    name="feather_creator",
    version="0.0.3",
    author="Timothée Frouté and Sergio Peignier",
    author_email="sergio.peignier@insa-lyon.fr",
    description="motif-TG .feather files creation and motifs-TF mapping from a list",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    install_requires = requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
)

print(setuptools.find_packages())



"""
feather_creator

This package allows creating and saving feather files, from a motif database 
and gene sequences. It also allows mapping motif-TF interactions from 
a list of specific transcription factors.
"""
__author__ = "Timothée Frouté and Sergio Peignier"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.0.1"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"
